package com.analogbit.puzzletools.client.slidingblocks.gui;

import com.analogbit.puzzletools.client.slidingblocks.state.Block;
import com.analogbit.puzzletools.client.util.Color;
import com.analogbit.puzzletools.client.util.DeepCloneable;
import com.google.gwt.event.dom.client.HasAllMouseHandlers;
import com.google.gwt.event.dom.client.MouseDownEvent;
import com.google.gwt.event.dom.client.MouseDownHandler;
import com.google.gwt.event.dom.client.MouseMoveEvent;
import com.google.gwt.event.dom.client.MouseMoveHandler;
import com.google.gwt.event.dom.client.MouseOutEvent;
import com.google.gwt.event.dom.client.MouseOutHandler;
import com.google.gwt.event.dom.client.MouseOverEvent;
import com.google.gwt.event.dom.client.MouseOverHandler;
import com.google.gwt.event.dom.client.MouseUpEvent;
import com.google.gwt.event.dom.client.MouseUpHandler;
import com.google.gwt.event.dom.client.MouseWheelEvent;
import com.google.gwt.event.dom.client.MouseWheelHandler;
import com.google.gwt.event.shared.HandlerRegistration;
import com.google.gwt.user.client.Event;
import com.google.gwt.user.client.ui.AbsolutePanel;
import com.google.gwt.user.client.ui.Image;

/**
 * Panel that displays a single block, which may be comprised of many tiles.
 */
public class GridBlock extends AbsolutePanel implements DeepCloneable, HasAllMouseHandlers {
	
	/** The size of a tile, images must be this dimension too. */
	public static final int CELL_SIZE = 52;
	
	/** Location to look for the image manipulator script. */
	protected static final String BASE_URL = "http://img.analogbit.com/hsv.php?";
	/** Image name of a normal block */ 
	protected static final String IMG_NAME = "block.gif";
	/** Image name of a locked block */
	protected static final String LOCKEDIMG_NAME = "lockblock.gif";
	
	/** The image can be highlighted */
	public enum MODE {
		HIGHLIGHTED, NORMAL
	}
	
	private MODE mode;
	private Block sourceBlock;
	//private HandlerManager hanMan;
	
	public GridBlock(Block sourceBlock) {
		this.sourceBlock = sourceBlock;
		updateBlock();
		
		this.sinkEvents(Event.MOUSEEVENTS);
	}

	/**
	 * Redraws the block on the panel.
	 */
	public void updateBlock() {
		int newWidth = CELL_SIZE * sourceBlock.getWidth();
		int newHeight = CELL_SIZE * sourceBlock.getHeight();
		this.setPixelSize(newWidth, newHeight);
		
		String cellURL =  buildUrl(sourceBlock.getColor(), sourceBlock.isImmobile(), false);
		String highlightURL = buildUrl(sourceBlock.getColor(), sourceBlock.isImmobile(), true);
		Image.prefetch(highlightURL);

		// this crap looks at the pile of bits representing block tiles. If it
		// set, it adds a graphic for the tile. this should probably be 
		// abstracted
		int bitIndex = 0;
		boolean blockBits[] = sourceBlock.getBlockBits();
		for (int y=0; y<sourceBlock.getHeight(); y++) {
			int yPixel = CELL_SIZE * y;
			for (int x=0; x<sourceBlock.getWidth(); x++) {
				if (blockBits[bitIndex]) {
					int xPixel = CELL_SIZE * x;
					Image normalBlock = new Image(cellURL); 
					this.add(normalBlock, xPixel, yPixel);
				}
				bitIndex++;
			}
		}	
	}
	
	/**
	 * Retrieves the block object that this panel displays.
	 * @return the block object that this panel displays.
	 */
	public Block getSourceBlock() {
		return sourceBlock;
	}
	
	/**
	 * Sets the block as highlighted or not.
	 * @param mode Mode to change to
	 */
	public void setMode(MODE mode) {
		if (this.mode == mode) {
			return;
		}
		
		boolean highlighted = mode == MODE.HIGHLIGHTED; 
		String url = buildUrl(sourceBlock.getColor(), sourceBlock.isImmobile(), highlighted);
		for (int i=0; i<this.getWidgetCount(); i++) {
			Image img = (Image)this.getWidget(i);
			img.setUrl(url);
		}
		this.mode = mode;
	}
	
	
	/**
	 * Takes parameters and builds a URL to display the appropriate tile.
	 * @param color Color for the block
	 * @param locked To display the lock icon, set this to true.
	 * @param highlight To display as a highlighted block set this to true
	 * @return The URL of the desired tile.
	 */
	public static String buildUrl(Color color, boolean locked, boolean highlight) {
		double [] hsv = new double[3];
		color.getHueSatValue(hsv);
		
		if (highlight) {
			hsv[2] = Math.sqrt(hsv[2]) * 1.2;
		}
	
		String url = BASE_URL;
		url += "img=" + (locked?LOCKEDIMG_NAME:IMG_NAME);
		//hue in image magick is weird
		url += "&hue=" + (int)(hsv[0]*200 + 100);
		if (locked) {
			url += "&sat=" + 75;
		}
		url += "&value=" + (int)(hsv[2]*100);
		//System.out.println("Color: " + color + " to url: " + url);
		return url;
	}
	
	/**
	 * Fetches all the possible highlight and icon combinations for a tile of 
	 * a specific color.
	 * @param color The color of the block
	 */
	public static void preloadColor(Color color) {
		Image.prefetch(buildUrl(color, false, false));
		Image.prefetch(buildUrl(color, false, true));
		Image.prefetch(buildUrl(color, true, false));
		Image.prefetch(buildUrl(color, true, true));
	}

	public Object deepClone() {
		return new GridBlock((Block)sourceBlock.deepClone());
	}

	public HandlerRegistration addMouseDownHandler(MouseDownHandler handler) {
		return addDomHandler(handler, MouseDownEvent.getType());
	}

	public HandlerRegistration addMouseUpHandler(MouseUpHandler handler) {
		return addDomHandler(handler, MouseUpEvent.getType());
	}

	public HandlerRegistration addMouseOutHandler(MouseOutHandler handler) {
		return addDomHandler(handler, MouseOutEvent.getType());
	}

	public HandlerRegistration addMouseOverHandler(MouseOverHandler handler) {
		return addDomHandler(handler, MouseOverEvent.getType());
	}

	public HandlerRegistration addMouseMoveHandler(MouseMoveHandler handler) {
		return addDomHandler(handler, MouseMoveEvent.getType());
	}

	public HandlerRegistration addMouseWheelHandler(MouseWheelHandler handler) {
		return addDomHandler(handler, MouseWheelEvent.getType());
	}
}
