package com.analogbit.puzzletools.client.slidingblocks.gui;

import java.util.ArrayList;

import com.allen_sauer.gwt.dnd.client.PickupDragController;
import com.analogbit.puzzletools.client.slidingblocks.state.Block;
import com.analogbit.puzzletools.client.slidingblocks.state.Grid;
import com.analogbit.puzzletools.client.slidingblocks.state.GridListener;
import com.analogbit.puzzletools.client.slidingblocks.state.GridWithListeners;
import com.google.gwt.dom.client.NativeEvent;
import com.google.gwt.event.shared.HandlerRegistration;
import com.google.gwt.user.client.DOM;
import com.google.gwt.user.client.Event;
import com.google.gwt.user.client.Event.NativePreviewEvent;
import com.google.gwt.user.client.Event.NativePreviewHandler;
import com.google.gwt.user.client.ui.AbsolutePanel;

/**
 * Main subclass that displays the grid and keeps track of what the mouse is over. 
 */
public class GridDisplay extends AbsolutePanel implements GridListener, NativePreviewHandler {
	private ArrayList<GridDisplayListener> listeners = new ArrayList<GridDisplayListener>();
	private int mouseOverX;
	private int mouseOverY;
	private int mouseDownX;
	private int mouseDownY;
	private boolean dragged = false;
	private ArrayList<GridBlock> blocks;
	private Grid gridDisplaying;
	private GridBlock blockHighlighted;
	private PickupDragController dragController;
	private HandlerRegistration previewHandler;

	public GridDisplay() {
		this(null, true);
	}
	
	/**
	 * Construct with a specific drag controller.
	 */
	public GridDisplay(AbsolutePanel dragBoundary, boolean allowDroppingOnBoundaryPanel) {
		this.addStyleName("gridDisplay");
		this.sinkEvents(Event.MOUSEEVENTS | Event.ONCONTEXTMENU);
		
		if (dragBoundary == null) {
			dragBoundary = this;
		}
		
		dragController = new PickupDragController(dragBoundary, allowDroppingOnBoundaryPanel) {
			@Override
			public void dragStart() {
				dragged = true;
				super.dragStart();
			}
		};
		dragController.setBehaviorDragStartSensitivity(6);
		
		blocks = new ArrayList<GridBlock>();
		mouseDownX = -1;
		mouseDownY = -1;
		mouseOverX = -1;
		mouseOverY = -1;

		blockHighlighted = null;
	}
	
	/**
	 * Get last known X location of the mouse.
	 * @return X coordinate of last grid cell the mouse was at.
	 */
	public int getCellMouseOverX() {
		return mouseOverX;
	}

	/**
	 * Get last known Y location of the mouse.
	 * @return Y coordinate of last grid cell the mouse was at.
	 */
	public int getCellMouseOverY() {
		return mouseOverY;
	}

	/**
	 * Get the X location of the mouse down cell.
	 * @return the X location of the mouse down cell.
	 */
	public int getCellMouseDownX() {
		return mouseDownX;
	}
	
	/**
	 * Get the Y location of the mouse down cell.
	 * @return the Y location of the mouse down cell.
	 */
	public int getCellMouseDownY() {
		return mouseDownY;
	}

	/**
	 * Every GridDisplay has a PickupDragController.
	 * @return The drag controller for this grid display.
	 */
	public PickupDragController getDragController() {
		return dragController;
	}
	
	/**
	 * Change what grid we are showing.
	 * @param g The grid to display.
	 */
	public void setGridDisplaying(Grid g) {
		if (gridDisplaying != null) {
			if (gridDisplaying instanceof GridWithListeners) {
				GridWithListeners gwl = (GridWithListeners)gridDisplaying;
				gwl.removeGridListener(this);
			}
		}
		
		this.setPixelSize(GridBlock.CELL_SIZE * g.getWidth(), GridBlock.CELL_SIZE * g.getHeight());
		this.setPixelSize(GridBlock.CELL_SIZE * g.getWidth(), GridBlock.CELL_SIZE * g.getHeight());
		this.clear();
		blockHighlighted = null;
		
		generateBlocks(g);
		gridDisplaying = g;

		highlight(getGridBlockAt(mouseOverX, mouseOverY));

		//if the grid can support listeners, add this panel so we know when 
		//to redraw
		if (g instanceof GridWithListeners) {
			GridWithListeners gwl = (GridWithListeners)g;
			gwl.addGridListener(this);
		}
	}
	
	private void generateBlocks(Grid g) {
		ArrayList<Block> gBlocks = g.getBlocks();
		blocks.clear();
		for (Block block : gBlocks) {
			onBlockAdded(block);
		}
	}
	
	public void addGridListener(GridDisplayListener listener) {
		listeners.add(listener);
	}
	
	public boolean removeGridListener(GridDisplayListener listener) {
		return listeners.remove(listener);
	}

	/**
	 * Updates the last known location of the mouse.
	 * @param event The event to query for mouse location.
	 */
	private void updateMouseCoords(Event event) {
		int x = event.getClientX();
		x -= this.getAbsoluteLeft();
		x += getScrollXOffset();
		x /= GridBlock.CELL_SIZE;

		int y =event.getClientY();
		y -= this.getAbsoluteTop();
		y += getScrollYOffset();
		y /= GridBlock.CELL_SIZE;
		
		mouseOverX = x;
		mouseOverY = y;
	}
	
	//to correct for scrolling
    private static native int getScrollYOffset() /*-{
    	var scrOfY = 0;

        if( typeof( $wnd.pageYOffset ) == "number" ){
        	scrOfY = $wnd.pageYOffset;
        } else if( $doc.body && ( $doc.body.scrollLeft || $doc.body.scrollTop) ){
        	scrOfY = $doc.body.scrollTop;
        } else {
        	scrOfY = $doc.documentElement.scrollTop;
        }
        return scrOfY;
    }-*/;
    
    //to correct for scrolling
    private static native int getScrollXOffset() /*-{
    	var scrOfX = 0;
  
    	if( typeof( $wnd.pageXOffset ) == "number" ){
            scrOfX = $wnd.pageXOffset;
    	} else if( $doc.body && ( $doc.body.scrollLeft || $doc.body.scrollTop) ){
            scrOfX = $doc.body.scrollLeft;
		} else {
            scrOfX = $doc.documentElement.scrollLeft;
    	}
    	return scrOfX;
  	}-*/; 

	//this is a huge hack because dragging eats mouse-up events
	@Override
	public void onPreviewNativeEvent(NativePreviewEvent event) {
		if (event.getTypeInt() ==  Event.ONMOUSEUP) {
			NativeEvent nativeEvent = event.getNativeEvent();			
			if (nativeEvent.getButton() == NativeEvent.BUTTON_LEFT) {
				onLeftButtonUp(nativeEvent.getClientX(), nativeEvent.getClientY());
			}
		}
	}
    
    //this function exists so that we can use the handler hack to detect a mouse up
    public void onLeftButtonUp(int clientX, int clientY) {
    	if (mouseDownX != -1) { //only handle the mouse up if we think the mosue was down
			mouseDownX = -1;
			mouseDownY = -1;
			 
			fireGridMouseUp(!dragged, mouseOverX, mouseOverY, clientX, clientY);
    	}
    	if (previewHandler != null) {
    		previewHandler.removeHandler();
    	}
    	
    	highlight(getGridBlockAt(mouseOverX, mouseOverY));
    }
    
	public void onBrowserEvent(Event event) {
		switch (DOM.eventGetType(event)) {
		case Event.ONMOUSEUP:
			switch (event.getButton()) {
			case Event.BUTTON_LEFT:
				updateMouseCoords(event);
				onLeftButtonUp(event.getClientX(), event.getClientY());						
				break;				
			}
			break;
		case Event.ONMOUSEDOWN:
			switch (event.getButton()) {
			case Event.BUTTON_LEFT:
				updateMouseCoords(event);
				mouseDownX = mouseOverX;
				mouseDownY = mouseOverY;
				dragged = false;
				
				Block blockDown = gridDisplaying.getBlockAt(mouseOverX, mouseOverY);
			
				fireGridMouseDown(blockDown, mouseDownX, mouseDownY, event.getClientX(), event.getClientY());
				//disallow drag image behavior
				event.preventDefault();
				
				//hack to recieve a mouse-up
				previewHandler = Event.addNativePreviewHandler(this);
				break;
			}
			break;
		case Event.ONCONTEXTMENU:
			event.preventDefault(); 
			updateMouseCoords(event);
			Block blockDown = gridDisplaying.getBlockAt(mouseOverX, mouseOverY);
			fireGridContextMenu(blockDown, mouseOverX, mouseOverY, event.getClientX(), event.getClientY());
			break;
		case Event.ONMOUSEMOVE:
			updateMouseCoords(event);
			GridBlock blockOver = getGridBlockAt(mouseOverX, mouseOverY);
			fireGridMouseMove(mouseOverX, mouseOverY, event.getClientX(), event.getClientY());
			highlight(blockOver);
			break;
		case Event.ONMOUSEOUT:
			highlight(null);
			break;
		}
		super.onBrowserEvent(event);
	}

	/**
	 * Converts from cell coordinates to grid list index.
	 * @param x X location of cell
	 * @param y Y location of cell
	 * @return The grid list index of the block, or -1 if no block present 
	 * in the cell.
	 */
	protected int getGridBlockIndexAt(int x, int y) {
		for (int i=0; i<blocks.size(); i++) {
			GridBlock block = blocks.get(i); 
			if (block.getSourceBlock().hasCellAt(x, y)) {
				return i; 
			}
		}
		return -1;
	}
	
	/**
	 * Retrieves the grid block at the specified location
	 * @param x X location of cell
	 * @param y Y location of cell
	 * @return Corresponding grid block, or null.
	 */
	protected GridBlock getGridBlockAt(int x, int y) {
		int index = getGridBlockIndexAt(x, y);
		if (index == -1) {
			return null;
		}
		return blocks.get(index);
	}
	
	/**
	 * Finds the grid list index of the specified grid block.
	 * @param block The block to get the index of
	 * @return Grid list index of the specified grid block, or -1 if it is not in the list. 
	 */
	protected int getGridBlockIndex(Block block) {
		for (int i=0; i<blocks.size(); i++) {
			GridBlock gridBlock = blocks.get(i); 
			if (gridBlock.getSourceBlock() == block) {
				return i; 
			}
		}
		return -1;
	}

	/**
	 * Gets the GridBlock object of a state Block.
	 * @param block The block from the state object.
	 * @return A gridblock that is being used to display the state block.
	 */
	protected GridBlock getGridBlock(Block block) {
		for (int i=0; i<blocks.size(); i++) {
			GridBlock gridBlock = blocks.get(i); 
			if (gridBlock.getSourceBlock() == block) {
				return gridBlock; 
			}
		}
		return null;
	}
	
	/**
	 * Highlights a grid block. Unhighlights the previous highlighted block.
	 * @param gridBlock The grid block to highlight, or null to unhighlight the current.
	 */
	protected void highlight(GridBlock gridBlock) {
		if (blockHighlighted != gridBlock) {
			
			if (blockHighlighted != null) {
				blockHighlighted.setMode(GridBlock.MODE.NORMAL);
				//move it back to the bottom
				blockHighlighted.getElement().getStyle().setProperty("zIndex", "0");
				blockHighlighted = null;
			}
			
			if (gridBlock == null || !isHighlightable(gridBlock)) {
				return;
			}
			
			blockHighlighted = gridBlock;
			
			//moves it to the top so we can drag it
			if (blockHighlighted != null) {
				//move it to the top so we grab the right piece
				gridBlock.getElement().getStyle().setProperty("zIndex", "1");
				gridBlock.setMode(GridBlock.MODE.HIGHLIGHTED);
			}
		}
	}

	private void fireGridMouseMove(int gridX, int gridY, int clientX, int clientY) {
		for (GridDisplayListener gridListener : listeners) {
			gridListener.onGridMouseMove(gridX, gridY, clientX, clientY);
		}
	}
	
	private void fireGridMouseUp(boolean clicked, int gridX, int gridY, int clientX, int clientY) {
		for (GridDisplayListener gridListener : listeners) {
			gridListener.onGridMouseUp(clicked, gridX, gridY, clientX, clientY);
		}
	}

	private void fireGridMouseDown(Block blockDown, int gridX, int gridY, int clientX, int clientY) {
		for (GridDisplayListener gridListener : listeners) {
			gridListener.onGridMouseDown(blockDown, gridX, gridY, clientX, clientY);
		}
	}

	private void fireGridContextMenu(Block blockDown, int gridX, int gridY, int clientX, int clientY) {
		for (GridDisplayListener gridListener : listeners) {
			gridListener.onGridContextMenu(blockDown, gridX, gridY, clientX, clientY);
		}
	}
	
	public void updateGridBlockPosition(GridBlock gridBlock) {
		Block block = gridBlock.getSourceBlock();
		int left = block.getLeft()*GridBlock.CELL_SIZE;
		int top = block.getTop()*GridBlock.CELL_SIZE;
		try {
			this.setWidgetPosition(gridBlock, left, top);
		} catch (IllegalArgumentException e) {
			//this can happen from drag and dropping
		}
	}
	
	
	public void onBlockAdded(Block block) {
		GridBlock gridBlock = new GridBlock(block);
		this.add(gridBlock);
		updateGridBlockPosition(gridBlock);
		blocks.add(gridBlock);
		
		if (isDraggable(gridBlock)) {
			dragController.makeDraggable(gridBlock);
		}
	}

	public void onBlockMoved(Block block, int fromX, int fromY) {
		GridBlock gridBlock = getGridBlock(block);
		if (gridBlock != null) {
			updateGridBlockPosition(gridBlock);
		}
	}
	
	public void onBlockChanged(Block block) {
		GridBlock gridBlock = getGridBlock(block);
		if (gridBlock != null) {
			gridBlock.updateBlock();
		}
	}

	public void onBlockRemoved(Block block) {
		int blockIndex = getGridBlockIndex(block);
		if (blockIndex != -1) {
			this.remove(blocks.remove(blockIndex));
		}
	}
	
	/**
	 * This method is meant to be overridden. And is used by the base class
	 * to determine if a block should be draggable.
	 * @param block The block to determine if it should be draggable.
	 * @return true if the block is draggable.
	 */
	public boolean isDraggable(GridBlock block) {
		return false;
	}

	/**
	 * This method is meant to be overridden. And is used by the base class
	 * to determine if a block should be highlightable.
	 * @param block The block to determine if it should be highlightable.
	 * @return true if the block is highlightable.
	 */
	public boolean isHighlightable(GridBlock block) {
		return true;
	}

}
