package com.analogbit.puzzletools.client.slidingblocks.state;

import com.analogbit.puzzletools.client.util.Color;
import com.analogbit.puzzletools.client.util.DeepCloneable;

public final class Block implements DeepCloneable {

	private int x;
	private int y;
	private BlockShape blockShape;
	private Color color;
	private boolean immobile;
	
	public Block(int x, int y, BlockShape blockShape, Color color, boolean immobile) {
		this.x = x;
		this.y = y;
		this.blockShape = blockShape;
		this.color = color;
		this.immobile = immobile;
	}

	public Block(Block block) {
		this(block.x, block.y, block.blockShape, block.color, block.immobile);
	}

	public boolean isImmobile() {
		return immobile;
	}

	protected void setImmobile(boolean immobile) {
		this.immobile = immobile;
	}

	protected void setColor(Color color) {
		this.color = color;
	}

	public boolean[] getBlockBits() {
		return blockShape.getBlockBits();
	}

	public int getHeight() {
		return blockShape.getHeight();
	}

	public int getWidth() {
		return blockShape.getWidth();
	}

	protected int getX() {
		return x;
	}

	protected void setX(int x) {
		this.x = x;
	}

	protected int getY() {
		return y;
	}

	protected void setY(int y) {
		this.y = y;
	}

	public int getTop() {
		return getY();
	}

	public int getBottom() {
		return getY() + getHeight();
	}
	
	public int getLeft() {
		return getX();
	}
	
	public int getRight() {
		return getX() + getWidth();
	}
	
	public BlockShape getBlockShape() {
		return blockShape;
	}

	public void setBlockShape(BlockShape blockShape) {
		this.blockShape = blockShape;
	}
	
	@Override
	public int hashCode() {
		return (x*31 * y*13) ^ blockShape.hashCode();
	}

	@Override
	public boolean equals(Object obj) {
		Block other = (Block)obj;
		return this.x == other.x && this.y == other.y && this.blockShape.equalsNoCast(other.blockShape);
	}
	
	public boolean equalsNoCast(Block other) {
		return this.x == other.x && this.y == other.y && this.blockShape.equalsNoCast(other.blockShape);
	}

	public Object deepClone() {
		return new Block(this);
	}

	public Block cloneShifted(int x, int y) {
		return new Block(getX() + x, getY() + y, getBlockShape(), getColor(), isImmobile());
	}

	public Block cloneWithNewPos(int x, int y) {
		return new Block(x, y, getBlockShape(), getColor(), isImmobile());
	}
	
	public static boolean intersects(Block block1, Block block2) {
		if (block1.getRight() <= block2.getLeft() ||
			block1.getBottom() <= block2.getTop()) {
			return false;
		}
		
		if (block2.getRight() <= block1.getLeft() ||
			block2.getBottom() <= block1.getTop()) {
			return false;
		}
		
		
		int overlapLeft = Math.max(block1.getLeft(), block2.getLeft());
		int overlapRight = Math.min(block1.getRight(), block2.getRight());
		int overlapTop = Math.max(block1.getTop(), block2.getTop());
		int overlapBottom = Math.min(block1.getBottom(), block2.getBottom());
		
		for (int y = overlapTop; y<overlapBottom; y++) {
			for (int x = overlapLeft; x<overlapRight; x++) {
				if (block1.hasCellAt(x, y) && block2.hasCellAt(x, y)) {
					return true;
				}
			}
		}
		return false;
	}

	public Color getColor() {
		return this.color;
	}

	public boolean hasCellAt(int x, int y) {
		if (getLeft() > x || getRight() <= x || getTop() > y || getBottom() <= y) {
			return false;
		} else {
			return blockShape.isCellPresent(x - getLeft(), y-getTop());
		}
	}

}
