package com.analogbit.puzzletools.client.slidingblocks.state;

/**
 * A block on the grid. This class is immutable.
 */
public final class BlockShape {
	private final int width;
	private final int height;
	
	/**
	 *  This array is to be held onto, but if it you need to modify it, COPY it first.
	 */
	private final boolean [] blockBits;

	/**
	 * Creates a new block
	 * @param width Width of the block
	 * @param height Height of the block
	 * @param blockBits True represents a block square is present, array values are row-major. Due to memory and speed concerns, the blockBits are not cloned, so be careful.
	 */
	public BlockShape(int width, int height, boolean [] blockBits) {
		assert(blockBits.length == width*height);
		this.width = width;
		this.height = height;
		this.blockBits = blockBits;
	}
	
	/**
	 * Copy constructor
	 * @param copy Block to copy
	 */
	public BlockShape(BlockShape copy) {
		this(copy.width, copy.height, copy.blockBits);
	}

	public int getWidth() {
		return width;
	}

	public int getHeight() {
		return height;
	}
	
	public boolean [] getBlockBits() {
		return blockBits;
	}

	@Override
	public int hashCode() {
		return (257 * height + width) ^ blockBitsHash(blockBits);
	}

	@Override
	public boolean equals(Object obj) {
		BlockShape b = (BlockShape)obj;
		return this.width == b.width && this.height == b.height && blockBitsEquals(this.blockBits, b.blockBits);
	}
	
	/**
	 * Checks to see if two blockBits are equal
	 * @param b1 First array of block bits
	 * @param b2 Second array of block bits
	 * @return true if equal
	 */
	private static boolean blockBitsEquals(boolean [] b1, boolean [] b2) {
		assert(b1.length == b2.length);
		for (int i=0; i<b1.length; i++) {
			if (b1[i] != b2[i]) {
				return false;
			}
		}
		return true;
	}
	
	private static int blockBitsHash(boolean [] bits) {
		int hash = bits.length*67;
		for (int i=0; i<bits.length; i++) {
			if (bits[i]) {
				hash ^= (i+1) << (i&15);
			}
		}
		return hash;
	}

	public boolean equalsNoCast(BlockShape blockShape) {
		return this.width == blockShape.width && this.height == blockShape.height && blockBitsEquals(this.blockBits, blockShape.blockBits);
	}

	public boolean isCellPresent(int x, int y) {
		return blockBits[x + width*y];
	}

	
	/*
	public Iterator<Point> iterateCells() {
		return new CellIterator();
	}

	public class CellIterator implements Iterator<Point> {

		private int x;
		private int y;
		private Point next = null;
		
		public CellIterator() {
			x = 0;
			y = 0;
		}
		
		public boolean hasNext() {
			if (next == null) {
				for (; x < getWidth(); x++) {
					for (; y < getHeight(); y++) {
						if (getBlockShape().isCellPresent(x, y)) {
							next = new Point(getLeft() + x, getTop() + y);
						}
					}
				}
			}
			
			return next != null;
		}

		public Point next() {
			Point result = next;
			next = null;
			return result;
		}

		public void remove() {
			throw new UnsupportedOperationException();
		}
		
	}
	*/
}